"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isDefineConfigFunction = exports.insertValueInJSString = exports.insertValuesInJavaScript = exports.insertValuesInConfigFile = void 0;
const parser_1 = require("@babel/parser");
const recast_1 = require("recast");
const fs = __importStar(require("fs-extra"));
const utils_1 = require("../../utils");
async function insertValuesInConfigFile(filePath, obj = {}) {
    await insertValuesInJavaScript(filePath, obj);
    return true;
}
exports.insertValuesInConfigFile = insertValuesInConfigFile;
async function insertValuesInJavaScript(filePath, obj) {
    const fileContents = await fs.readFile(filePath, { encoding: 'utf8' });
    let finalCode = await insertValueInJSString(fileContents, obj);
    const prettifiedCode = await (0, utils_1.prettifyCode)(finalCode);
    if (prettifiedCode) {
        finalCode = prettifiedCode;
    }
    await fs.writeFile(filePath, finalCode);
}
exports.insertValuesInJavaScript = insertValuesInJavaScript;
async function insertValueInJSString(fileContents, obj) {
    const ast = (0, parser_1.parse)(fileContents, { plugins: ['typescript'], sourceType: 'module' });
    let objectLiteralNode;
    function handleExport(nodePath) {
        if (nodePath.node.type === 'CallExpression'
            && nodePath.node.callee.type === 'Identifier') {
            const functionName = nodePath.node.callee.name;
            if (isDefineConfigFunction(ast, functionName)) {
                return handleExport(nodePath.get('arguments', 0));
            }
        }
        if (nodePath.node.type === 'ObjectExpression' && !nodePath.node.properties.find((prop) => prop.type !== 'ObjectProperty')) {
            objectLiteralNode = nodePath.node;
            return;
        }
        throw new Error('Cypress was unable to add/update values in your configuration file.');
    }
    (0, recast_1.visit)(ast, {
        visitAssignmentExpression(nodePath) {
            if (nodePath.node.left.type === 'MemberExpression') {
                if (nodePath.node.left.object.type === 'Identifier' && nodePath.node.left.object.name === 'module'
                    && nodePath.node.left.property.type === 'Identifier' && nodePath.node.left.property.name === 'exports') {
                    handleExport(nodePath.get('right'));
                }
            }
            return false;
        },
        visitExportDefaultDeclaration(nodePath) {
            handleExport(nodePath.get('declaration'));
            return false;
        },
    });
    const splicers = [];
    if (!objectLiteralNode) {
        // if the export is no object litteral
        throw new Error('Cypress was unable to add/update values in your configuration file.');
    }
    setRootKeysSplicers(splicers, obj, objectLiteralNode, '  ');
    setSubKeysSplicers(splicers, obj, objectLiteralNode, '  ', '  ');
    // sort splicers to keep the order of the original file
    const sortedSplicers = splicers.sort((a, b) => a.start === b.start ? 0 : a.start > b.start ? 1 : -1);
    if (!sortedSplicers.length)
        return fileContents;
    let nextStartingIndex = 0;
    let resultCode = '';
    sortedSplicers.forEach((splicer) => {
        resultCode += fileContents.slice(nextStartingIndex, splicer.start) + splicer.replaceString;
        nextStartingIndex = splicer.end;
    });
    return resultCode + fileContents.slice(nextStartingIndex);
}
exports.insertValueInJSString = insertValueInJSString;
function isDefineConfigFunction(ast, functionName) {
    let value = false;
    (0, recast_1.visit)(ast, {
        visitVariableDeclarator(nodePath) {
            var _a, _b, _c;
            // if this is a require of cypress
            if (((_a = nodePath.node.init) === null || _a === void 0 ? void 0 : _a.type) === 'CallExpression'
                && nodePath.node.init.callee.type === 'Identifier'
                && nodePath.node.init.callee.name === 'require'
                && nodePath.node.init.arguments[0].type === 'StringLiteral'
                && nodePath.node.init.arguments[0].value === 'cypress') {
                if (((_b = nodePath.node.id) === null || _b === void 0 ? void 0 : _b.type) === 'ObjectPattern') {
                    const defineConfigFunctionNode = nodePath.node.id.properties.find((prop) => {
                        return prop.type === 'ObjectProperty'
                            && prop.key.type === 'Identifier'
                            && prop.key.name === 'defineConfig';
                    });
                    if (defineConfigFunctionNode) {
                        value = ((_c = defineConfigFunctionNode.value) === null || _c === void 0 ? void 0 : _c.name) === functionName;
                    }
                }
            }
            return false;
        },
        visitImportDeclaration(nodePath) {
            var _a, _b;
            if (nodePath.node.source.type === 'StringLiteral'
                && nodePath.node.source.value === 'cypress') {
                const defineConfigFunctionNode = (_a = nodePath.node.specifiers) === null || _a === void 0 ? void 0 : _a.find((specifier) => {
                    return specifier.type === 'ImportSpecifier'
                        && specifier.imported.type === 'Identifier'
                        && specifier.imported.name === 'defineConfig';
                });
                if (defineConfigFunctionNode) {
                    value = ((_b = defineConfigFunctionNode.local) === null || _b === void 0 ? void 0 : _b.name) === functionName;
                }
            }
            return false;
        },
    });
    return value;
}
exports.isDefineConfigFunction = isDefineConfigFunction;
function setRootKeysSplicers(splicers, obj, objectLiteralNode, lineStartSpacer) {
    const objectLiteralStartIndex = objectLiteralNode.start + 1;
    // add values
    const objKeys = Object.keys(obj).filter((key) => ['boolean', 'number', 'string'].includes(typeof obj[key]));
    // update values
    const keysToUpdate = objKeys.filter((key) => {
        return objectLiteralNode.properties.find((prop) => {
            return prop.type === 'ObjectProperty'
                && prop.key.type === 'Identifier'
                && prop.key.name === key;
        });
    });
    keysToUpdate.forEach((key) => {
        const propertyToUpdate = propertyFromKey(objectLiteralNode, key);
        if (propertyToUpdate) {
            setSplicerToUpdateProperty(splicers, propertyToUpdate, obj[key], key, obj);
        }
    });
    const keysToInsert = objKeys.filter((key) => !keysToUpdate.includes(key));
    if (keysToInsert.length) {
        const valuesInserted = `\n${lineStartSpacer}${keysToInsert.map((key) => `${key}: ${JSON.stringify(obj[key])},`).join(`\n${lineStartSpacer}`)}`;
        splicers.push({
            start: objectLiteralStartIndex,
            end: objectLiteralStartIndex,
            replaceString: valuesInserted,
        });
    }
}
function setSubKeysSplicers(splicers, obj, objectLiteralNode, lineStartSpacer, parentLineStartSpacer) {
    const objectLiteralStartIndex = objectLiteralNode.start + 1;
    const keysToUpdateWithObjects = [];
    const objSubkeys = Object.keys(obj).filter((key) => typeof obj[key] === 'object').reduce((acc, key) => {
        keysToUpdateWithObjects.push(key);
        Object.entries(obj[key]).forEach(([subkey, value]) => {
            if (['boolean', 'number', 'string'].includes(typeof value)) {
                acc.push({ parent: key, subkey });
            }
        });
        return acc;
    }, []);
    // add values where the parent key needs to be created
    const subkeysToInsertWithoutKey = objSubkeys.filter(({ parent }) => {
        return !objectLiteralNode.properties.find((prop) => {
            return prop.type === 'ObjectProperty'
                && prop.key.type === 'Identifier'
                && prop.key.name === parent;
        });
    });
    const keysToInsertForSubKeys = {};
    subkeysToInsertWithoutKey.forEach((keyTuple) => {
        const subkeyList = keysToInsertForSubKeys[keyTuple.parent] || [];
        subkeyList.push(keyTuple.subkey);
        keysToInsertForSubKeys[keyTuple.parent] = subkeyList;
    });
    let subvaluesInserted = '';
    for (const key in keysToInsertForSubKeys) {
        subvaluesInserted += `\n${parentLineStartSpacer}${key}: {`;
        keysToInsertForSubKeys[key].forEach((subkey) => {
            subvaluesInserted += `\n${parentLineStartSpacer}${lineStartSpacer}${subkey}: ${JSON.stringify(obj[key][subkey])},`;
        });
        subvaluesInserted += `\n${parentLineStartSpacer}},`;
    }
    if (subkeysToInsertWithoutKey.length) {
        splicers.push({
            start: objectLiteralStartIndex,
            end: objectLiteralStartIndex,
            replaceString: subvaluesInserted,
        });
    }
    // add/update values where parent key already exists
    keysToUpdateWithObjects.filter((parent) => {
        return objectLiteralNode.properties.find((prop) => {
            return prop.type === 'ObjectProperty'
                && prop.key.type === 'Identifier'
                && prop.key.name === parent;
        });
    }).forEach((key) => {
        const propertyToUpdate = propertyFromKey(objectLiteralNode, key);
        if ((propertyToUpdate === null || propertyToUpdate === void 0 ? void 0 : propertyToUpdate.value.type) === 'ObjectExpression') {
            setRootKeysSplicers(splicers, obj[key], propertyToUpdate.value, parentLineStartSpacer + lineStartSpacer);
        }
    });
}
function setSplicerToUpdateProperty(splicers, propertyToUpdate, updatedValue, key, obj) {
    if (propertyToUpdate && (isPrimitive(propertyToUpdate.value) || isUndefinedOrNull(propertyToUpdate.value))) {
        splicers.push({
            start: propertyToUpdate.value.start,
            end: propertyToUpdate.value.end,
            replaceString: JSON.stringify(updatedValue),
        });
    }
    else {
        throw new Error('Cypress was unable to add/update values in your configuration file.');
    }
}
function propertyFromKey(objectLiteralNode, key) {
    return objectLiteralNode === null || objectLiteralNode === void 0 ? void 0 : objectLiteralNode.properties.find((prop) => {
        return prop.type === 'ObjectProperty' && prop.key.type === 'Identifier' && prop.key.name === key;
    });
}
function isPrimitive(value) {
    return value.type === 'NumericLiteral' || value.type === 'StringLiteral' || value.type === 'BooleanLiteral';
}
function isUndefinedOrNull(value) {
    return value.type === 'Identifier' && ['undefined', 'null'].includes(value.name);
}
